// c++ src/test.cc --std=c++17 -Wall -Wextra -Wpedantic -o src/test -Ivendor/xxhash && src/test
#define XXH_STATIC_LINKING_ONLY 1
#define XXH_INLINE_ALL 1
#define XXH_PRIVATE_API 1
#include <cassert>
#include <cstring>
#include <type_traits>
#include <iostream>
#include <chrono>
#include <sys/random.h>
#include "xxhash.h"

inline uint64_t xxh3_64_seeded(const uint8_t *dat, size_t sz, uint64_t seed) {
  return ::XXH3_64bits_withSeed(dat, sz, seed);
}

inline uint64_t xxh3_64_concat(const uint8_t *dat, size_t sz, uint64_t seed) {
  uint64_t dat2[] = {
    ::XXH3_64bits(dat, sz),
    seed
  };
  return ::XXH3_64bits(dat2, sizeof(dat2));
}

inline uint64_t xxh3_64_xor(const uint8_t *dat, size_t sz, uint64_t seed) {
  if (XXH_likely(sz <= 16)) {
    // (Just for demonstration purposes; padding probably needs correcting)
    uint64_t data_words[] = { 0, 0 };
    memcpy(data_words, dat, sz);
    data_words[1] ^= seed;
    return ::XXH3_64bits(data_words, sizeof(data_words));
  } else {
    uint64_t dat2 = ::XXH3_64bits(dat, sz) ^ seed;
    return ::XXH3_64bits(&dat2, sizeof(dat2));
  }
}

inline XXH128_hash_t xxh3_128_seeded(const uint8_t *dat, size_t sz, XXH128_hash_t seed) {
  return ::XXH3_128bits_withSeed(dat, sz, seed.low64);
}

inline XXH128_hash_t xxh3_128_concat(const uint8_t *dat, size_t sz, XXH128_hash_t seed) {
  XXH128_hash_t dat2[] = {
    ::XXH3_128bits(dat, sz),
    seed
  };
  return ::XXH3_128bits(std::data(dat2), std::size(dat2));
}

inline XXH128_hash_t xxh3_128_xor(const uint8_t *dat, size_t sz, XXH128_hash_t seed) {
  XXH128_hash_t dat2;

  if (XXH_likely(sz <= 16)) {
    // (Just for demonstration purposes; padding probably needs correcting)
    memcpy(&dat2, dat, sz);
  } else {
    dat2 = ::XXH3_128bits(dat, sz);
  }

  dat2.high64 ^= seed.high64;
  dat2.low64 ^= seed.low64;
  return ::XXH3_128bits(&dat2, sizeof(dat2));
}

template<typename Fn>
double measure(Fn fn, uint8_t *dat, size_t sz) {
  using namespace std::chrono;
  using HashVal = decltype(fn(dat, sz, {}));

  HashVal accu{};
  auto t0 = high_resolution_clock::now();

  for (int i=0; i < 1000000; i++) {
    ::memcpy(dat, &accu, sizeof(accu));
    accu = fn(dat, sz, accu);
  }

  auto t1 = high_resolution_clock::now();
  return duration_cast<duration<double>>(t1 - t0).count() * 1.0e6;
}

int main() {
  uint8_t data[8192];
  assert(getrandom(data, sizeof(data), 0) == sizeof(data));

#define MEAS(len, sz) { \
  std::cout << len << " " << sz << " " << std::flush; \
\
  auto base = measure(xxh3_ ## len ## _seeded, data, sz); \
  std::cout << "1 (" << base << ") " << std::flush; \
\
  auto cat = measure(xxh3_ ## len ## _concat, data, sz); \
  std::cout << (base/cat) << " (" << cat << ") " << std::flush; \
\
  auto xoo = measure(xxh3_ ## len ## _xor, data, sz); \
  std::cout << (base/xoo) << " (" << xoo << ")" << std::endl; \
}
  MEAS(64, 1);
  MEAS(64, 2);
  MEAS(64, 4);
  MEAS(64, 8);
  MEAS(64, 16);
  MEAS(64, 32);
  MEAS(64, 128);
  MEAS(64, 8192);
  MEAS(128, 1);
  MEAS(128, 2);
  MEAS(128, 4);
  MEAS(128, 8);
  MEAS(128, 16);
  MEAS(128, 32);
  MEAS(128, 128);
  MEAS(128, 8192);
  return 0;
}
